@@ -1,5 +1,6 @@
 #define CHANGE_FONT_SIZE "change font size"
 #define GB_OUTPUT_TOGGLE "gb output toggle"
+#define KBM_TOGGLE "kbm toggle"
 
 
 #define GCIN_FONT_SIZE "gcin-font-size"
@@ -15,6 +16,7 @@
 #define GCIN_POP_UP_WIN "gcin-pop-up-win"
 #define GCIN_POP_UP_WIN_ABS_CORNER "gcin-pop-up-win-abs-corner"
 #define GCIN_INNER_FRAME "gcin-inner-frame"
+#define GCIN_INIT_IM_ENABLED "gcin-init-im-enabled"
 
 #define GCIN_IM_TOGGLE_KEYS "gcin-im-toggle-keys"
 #define DEFAULT_INPUT_METHOD "default-input-method"
@@ -23,7 +25,14 @@
 #define GCIN_FLAGS_IM_ENABLED "gcin-flags-im-enabled"
 #define GCIN_REMOTE_CLIENT "gcin-remote-client"
 #define GCIN_SHIFT_SPACE_ENG_FULL "gcin-shift-space-eng-full"
-
+#define GCIN_STATUS_TRAY "gcin-status-tray"
+#define GCIN_WIN_COLOR_FG "gcin-win-color-fg"
+#define GCIN_WIN_COLOR_BG "gcin-win-color-bg"
+#define GCIN_WIN_COLOR_USE "gcin-win-color-use"
+#define GCIN_CAPSLOCK_LOWER "gcin-capslock-lower"
+#define GCIN_ENG_PHRASE_ENABLED "gcin-eng-phrase-enabled"
+#define GCIN_WIN_SYM_CLICK_CLOSE "gcin-win-sym-click-close"
+#define GCIN_SEL_KEY_COLOR "gcin-sel-key-color"
 
 #define GTAB_DUP_SELECT_BELL "gtab-dup-select-bell"
 #define GTAB_SPACE_AUTO_FIRST "gtab-space-auto-first"
@@ -39,6 +48,10 @@
 #define GTAB_HIDE_ROW2 "gtab-hide-row2"
 #define GTAB_IN_ROW1 "gtab-in-row1"
 #define GTAB_CAPSLOCK_IN_ENG "gtab-capslock-in-eng"
+#define GTAB_VERTICAL_SELECT "gtab-vertical-select"
+#define GTAB_UNIQUE_AUTO_SEND "gtab-unique-auto-send"
+#define GTAB_QUE_WILD_CARD "gtab-que-wild-card"
+#define GTAB_EDIT_BUFFER "gtab-edit-buffer"
 
 #define PHO_SIMPLE_WIN "pho-simple-window" // common to both pho and tsin
 #define TSIN_PHRASE_PRE_SELECT "tsin-phrase-pre-select"
@@ -50,6 +63,8 @@
 #define TSIN_CURSOR_COLOR "tsin-cursor-color"
 #define TSIN_TONE_CHAR_INPUT "tsin-tone-char-input"
 #define TSIN_TAB_PHRASE_END "tsin-tab-phrase-end"
+#define TSIN_TAIL_SELECT_KEY "tsin-tail-select-key"
+#define TSIN_BUFFER_EDITING_MODE "tsin-buffer-editing-mode"
 
 #define PHO_HIDE_ROW2 "pho-hide-row2"
 #define PHO_IN_ROW1 "pho-in-row1"
@@ -58,6 +73,8 @@
 #define PHONETIC_KEYBOARD "phonetic-keyboard"
 #define PHONETIC_CHAR_DYNAMIC_SEQUENCE "phonetic-char-dynamic-sequence"
 #define PHONETIC_HUGE_TAB "phonetic-huge-tab"
+#define PHONETIC_SPEAK "phonetic-speak"
+#define PHONETIC_SPEAK_SEL "phonetic-speak-sel"
 
 
 extern int gcin_font_size, gcin_font_size_tsin_presel, gcin_font_size_symbol,
@@ -66,8 +83,8 @@ extern int gcin_font_size, gcin_font_size_tsin_presel, gcin_font_size_symbol,
            gcin_flags_im_enabled, gcin_remote_client, gtab_simple_win,
            gtab_disp_key_codes, gtab_disp_im_name, gcin_shift_space_eng_full,
            gtab_invalid_key_in, gtab_hide_row2, gtab_in_row1, gtab_capslock_in_eng,
-           pho_hide_row2,
-           pho_in_row1;
+           gcin_capslock_lower, pho_hide_row2, pho_in_row1,
+           gcin_eng_phrase_enabled, gtab_que_wild_card, gcin_win_sym_click_close;
 
 extern int default_input_method;
 extern int left_right_button_tips;
@@ -79,22 +96,31 @@ extern int gtab_pre_select;
 extern int gtab_press_full_auto_send;
 extern int gtab_disp_partial_match;
 extern int gtab_shift_phrase_key;
+extern int gtab_vertical_select, gtab_unique_auto_send, gtab_edit_buffer;
 extern int tsin_ec_toggle_key;
 extern int pho_simple_win, tsin_buffer_size;
 extern int gcin_input_style, gcin_root_x, gcin_root_y, gcin_pop_up_win, gcin_pop_up_win_abs_corner;
+extern int gcin_status_tray;
 
 extern int tsin_phrase_pre_select;
 extern int tsin_chinese_english_toggle_key;
-extern int tsin_tab_phrase_end;
+extern int tsin_tab_phrase_end, tsin_tail_select_key;
+extern int tsin_buffer_editing_mode;
 
 extern int phonetic_char_dynamic_sequence;
-extern int phonetic_huge_tab;
+extern int phonetic_huge_tab, phonetic_speak;
+extern char *phonetic_speak_sel;
 extern int tsin_space_opt, tsin_tone_char_input;
 
-extern char tsin_phrase_line_color[], tsin_cursor_color[], gcin_font_name[];
+extern char *tsin_phrase_line_color, *tsin_cursor_color, *gcin_font_name, *gcin_sel_key_color;
+extern char *gcin_win_color_fg, *gcin_win_color_bg;
+extern int gcin_win_color_use;
+extern int gcin_init_im_enabled;
 
 void get_gcin_user_fname(char *name, char fname[]);
-void get_gcin_conf_str(char *name, char rstr[], char *default_str);
+void get_gcin_conf_str(char *name, char **rstr, char *default_str);
+void get_gcin_conf_fstr(char *name, char rstr[], char *default_str);
 void save_gcin_conf_str(char *name, char *str);
 void save_gcin_conf_int(char *name, int val);
 void load_setttings();
+
