@@ -1,5 +1,5 @@
 /*
-	Copyright (C) 2004	Edward Der-Hua Liu, Hsin-Chu, Taiwan
+	Copyright (C) 2004-2008	Edward Der-Hua Liu, Hsin-Chu, Taiwan
 */
 #include <string.h>
 
@@ -16,6 +16,7 @@ extern char *pho_chars[];
 
 extern int ityp3_pho;
 extern u_char typ_pho[];
+gboolean typ_pho_empty();
 extern char inph[];
 
 extern u_short hash_pho[];
@@ -24,37 +25,25 @@ extern PHOKBM phkbm;
 extern char tsidxfname[64];
 extern int hashidx[TSIN_HASH_N];
 
-typedef struct {
-  phokey_t pho;
-  char ch[CH_SZ];
-  char och[CH_SZ];
-  char ph1ch[CH_SZ]; // char selected by 1st pho
-  u_char flag;
-  char psta; // phrase start index
-} CHPHO;
-
-enum {
-  FLAG_CHPHO_FIXED=1,    // user selected the char, so it should not be changed
-  FLAG_CHPHO_PHRASE_HEAD=2,
-  FLAG_CHPHO_PHRASE_VOID=4
-};
-
-static CHPHO chpho[MAX_PH_BF_EXT];
-static int c_idx, c_len, ph_sta=-1, ph_sta_last=-1;
+CHPHO chpho[MAX_PH_BF_EXT];
+int c_idx;
+extern int c_len;
+int ph_sta=-1, ph_sta_last=-1;  // phrase start
 static int sel_pho;
-static gboolean eng_ph=TRUE;  // english(FALSE) <-> pho(juyin, TRUE)
+gboolean eng_ph=TRUE;  // english(FALSE) <-> pho(juyin, TRUE)
 static int save_frm, save_to;
 static int current_page;
 static int startf;
 static gboolean full_match;
 static gboolean tsin_half_full;
+static gboolean tsin_buffer_editing = 0;
 
 typedef struct {
   phokey_t phokey[MAX_PHRASE_LEN];
   int phidx;
   char str[MAX_PHRASE_LEN*CH_SZ+1];
   int len;
-  char usecount;
+  usecount_t usecount;
 } PRE_SEL;
 
 static PRE_SEL pre_sel[10];
@@ -111,12 +100,14 @@ void drawcursor()
   last_cursor_idx = c_idx;
 
   if (c_idx == c_len) {
-    if (tsin_half_full || eng_ph) {
-      disp_char(c_idx,"  ");
-      set_cursor_tsin(c_idx);
-    } else {
-      disp_char(c_idx, " ");
-      set_cursor_tsin(c_idx);
+    if (!eng_ph) {
+      if (tsin_half_full) {
+        disp_char(c_idx,"  ");
+        set_cursor_tsin(c_idx);
+      } else {
+        disp_char(c_idx, " ");
+        set_cursor_tsin(c_idx);
+      }
     }
   }
   else {
@@ -145,17 +136,20 @@ static void chpho_extract(CHPHO *chph, int len, phokey_t *pho, char *ch, char *o
 void inc_pho_count(phokey_t key, int ch_idx);
 int ch_key_to_ch_pho_idx(phokey_t phkey, char *big5);
 void inc_dec_tsin_use_count(phokey_t *pho, char *ch, int prlen, int N);
+void lookup_gtabn(char *ch, char *);
 
 static void putbuf(int len)
 {
   u_char tt[CH_SZ * (MAX_PH_BF_EXT+1) + 1];
   int i,idx;
 
+//  dbg("putbuf:%d\n", len);
 
 #if 1
   // update phrase reference count
   if (len >= 2) {
     for(i=0; i < len; i++) {
+//      dbg("flag %d %x\n", i, chpho[i].flag);
       if (!BITON(chpho[i].flag, FLAG_CHPHO_PHRASE_HEAD)) {
         continue;
       }
@@ -177,7 +171,7 @@ static void putbuf(int len)
       puts("");
 #endif
 
-      if (phrlen < 2)
+      if (phrlen < 1)
         continue;
 
       phokey_t pho[MAX_PHRASE_LEN];
@@ -186,8 +180,11 @@ static void putbuf(int len)
 
       chpho_extract(&chpho[i], phrlen, pho, ch, och);
 
-      if (chpho[i].flag & FLAG_CHPHO_PHRASE_VOID)
+      if (chpho[i].flag & FLAG_CHPHO_PHRASE_VOID) {
+#if 0
         inc_dec_tsin_use_count(pho, och, phrlen, TRUE);
+#endif
+      }
       else
         inc_dec_tsin_use_count(pho, ch, phrlen, FALSE);
     }
@@ -209,6 +206,7 @@ static void putbuf(int len)
 
   tt[idx]=0;
   send_text(tt);
+  lookup_gtabn(tt, NULL);
 }
 
 
@@ -218,31 +216,28 @@ static void prbuf()
 {
   int i;
 
-//  dbg("prbuf\n");
-
   for(i=0; i < c_len; i++) {
     disp_char_chbuf(i);
   }
 
-  if (c_len < MAX_PH_BF_EXT)
-    hide_char(c_len);
-
-  if (c_len+1 < MAX_PH_BF_EXT)
-    hide_char(c_len+1);
-
   drawcursor();
 }
 
 
-static int orig_disp_in_area_x = -1;
 void disp_tsin_pho(int index, char *pho);
 
 static void disp_in_area_pho_tsin()
 {
   int i;
 
-  for(i=0;i<4;i++) {
-    disp_tsin_pho(i, &pho_chars[i][typ_pho[i] * PHO_CHAR_LEN]);
+  if (pin_juyin) {
+    for(i=0;i<6;i++) {
+      disp_tsin_pho(i, &inph[i]);
+    }
+  } else {
+    for(i=0;i<4;i++) {
+      disp_tsin_pho(i, &pho_chars[i][typ_pho[i] * PHO_CHAR_LEN]);
+    }
   }
 }
 
@@ -250,13 +245,10 @@ static void disp_in_area_pho_tsin()
 void clear_chars_all();
 
 
-static void clear_disp_ph_sta();
 static void clear_match()
 {
   ph_sta=-1;
-  clear_disp_ph_sta();
-//  bzero(psta, sizeof(psta));
-//  dbg("clear_match\n");
+  clear_tsin_line();
 }
 
 static void clr_ch_buf()
@@ -277,7 +269,7 @@ static void clear_ch_buf_sel_area()
   full_match = FALSE;
   clr_ch_buf();
   drawcursor();
-  clear_disp_ph_sta();
+  clear_tsin_line();
 }
 
 static void close_selection_win();
@@ -287,24 +279,29 @@ static void clear_tsin_buffer()
   clear_ch_buf_sel_area();
   close_selection_win();
   pre_selN = 0;
+  tsin_buffer_editing = 0; //buffer editing is finished
 }
 
 void clr_in_area_pho_tsin();
 void close_win_pho_near();
 void compact_win0_x();
 
-void tsin_reset_in_pho()
+void tsin_reset_in_pho0()
 {
-  clrin_pho_tsin();
-  prbuf();
+//  prbuf();
   clr_in_area_pho_tsin();
   close_selection_win();
   pre_selN = 0;
   drawcursor();
-
   close_win_pho_near();
 }
 
+void tsin_reset_in_pho()
+{
+  clrin_pho_tsin();
+  tsin_reset_in_pho0();
+}
+
 gboolean flush_tsin_buffer()
 {
   tsin_reset_in_pho();
@@ -328,14 +325,14 @@ void disp_tsin_eng_pho(int eng_pho);
 
 void show_stat()
 {
+#if TRAY_ENABLED
+  load_tray_icon();
+#endif
   disp_tsin_eng_pho(eng_ph);
 }
 
 void load_tsin_db();
 
-void load_tsin_entry(int idx, u_char *len, char *usecount, phokey_t *pho,
-                    u_char *ch);
-
 #if 0
 void nputs(u_char *s, u_char len)
 {
@@ -351,7 +348,7 @@ static void dump_tsidx(int i)
 {
   phokey_t pho[MAX_PHRASE_LEN];
   u_char ch[MAX_PHRASE_LEN*CH_SZ];
-  char usecount;
+  usecount_t usecount;
   u_char len;
 
   load_tsin_entry(i, &len, &usecount, pho, ch);
@@ -388,42 +385,35 @@ static void dump_tsidx_all()
 void load_tab_pho_file();
 void show_win0();
 
-void init_tab_pp(int usenow)
+void init_tab_pp(gboolean init)
 {
   if (!ch_pho)
     load_tab_pho_file();
 
-  if (phcount && gwin0) {
-disp_prom:
-    show_stat();
-    clear_ch_buf_sel_area();
-    if (!gcin_pop_up_win)
-      show_win0();
-
-    return;
+  if (!phcount) {
+    load_tsin_db();
+    if (init)
+      clr_ch_buf();
+//    show_win0();
   }
 
-  load_tsin_db();
+  show_stat();
+  if (init)
+    clear_ch_buf_sel_area();
 
-  clr_ch_buf();
-  show_win0();
-
-//  dump_tsidx_all();
-  goto disp_prom;
+  if (!gcin_pop_up_win)
+    show_win0();
 }
 
 gboolean save_phrase_to_db2(CHPHO *chph, int len);
 
 static void save_phrase()
 {
-  int tt, i;
+  int i;
   u_char len;
 
-  if (save_frm==save_to)
-    return;
-
   if (save_frm > save_to) {
-    tt=save_frm;
+    int tt=save_frm;
     save_frm=save_to;
     save_to=tt;
   }
@@ -432,7 +422,8 @@ static void save_phrase()
     save_to--;
 
   len= save_to- save_frm+ 1;
-  if (len < 2 || len > MAX_PHRASE_LEN)
+
+  if (len <= 0 || len > MAX_PHRASE_LEN)
     return;
 
   for(i=save_frm;i<=save_to;i++) {
@@ -460,6 +451,16 @@ static void save_phrase()
   return;
 }
 
+
+static void set_fixed(int idx, int len)
+{
+  int i;
+  for(i=idx; i < idx+len; i++) {
+    chpho[i].flag |= FLAG_CHPHO_FIXED;
+    chpho[i].flag &= ~FLAG_CHPHO_PHRASE_USER_HEAD;
+  }
+}
+
 #define PH_SHIFT_N (tsin_buffer_size - 1)
 
 static void shift_ins()
@@ -473,6 +474,12 @@ static void shift_ins()
    if (c_len >= PH_SHIFT_N) {
      int ofs;
 
+     // set it fixed so that it will not cause partial phrase in the beginning
+     int fixedlen = c_len - 10;
+     if (fixedlen <= 0)
+       fixedlen = 1;
+     set_fixed(0, fixedlen);
+
      ofs = 1;
      putbuf(ofs);
 
@@ -491,7 +498,7 @@ static void shift_ins()
        chpho[j+1] = chpho[j];
      }
 
-     chpho[c_len].ch[0]=' ';
+     init_chpho_i(c_len);
     /*    prbuf(); */
    }
 
@@ -503,6 +510,7 @@ static void put_b5_char(char *b5ch, phokey_t key)
 {
    shift_ins();
 
+   init_chpho_i(c_idx);
    bchcpy(chpho[c_idx].ch, b5ch);
    bchcpy(chpho[c_idx].och, b5ch);
    bchcpy(chpho[c_idx].ph1ch, b5ch);
@@ -517,12 +525,10 @@ static void put_b5_char(char *b5ch, phokey_t key)
    }
 }
 
-
-#define MAX_PHRASE_SEL_N (9)
+#define MAX_PHRASE_SEL_N 10
 
 static u_char selstr[MAX_PHRASE_SEL_N][MAX_PHRASE_LEN * CH_SZ];
 static u_char sellen[MAX_PHRASE_SEL_N];
-static int selidx[MAX_PHRASE_SEL_N];
 
 static u_short phrase_count;
 static u_short pho_count;
@@ -542,37 +548,38 @@ static gboolean chpho_eq_pho(int idx, phokey_t *phos, int len)
 static void get_sel_phrase()
 {
   int sti,edi,j;
-  phokey_t key, stk[MAX_PHRASE_LEN];
-  u_char len, mlen, stch[MAX_PHRASE_LEN * CH_SZ + 1];
+  u_char len, mlen;
+
+  phrase_count = 0;
 
   mlen=c_len-c_idx;
 
+  if (!mlen)
+    return;
+
   if (mlen > MAX_PHRASE_LEN)
     mlen=MAX_PHRASE_LEN;
 
-  key=chpho[c_idx].pho;
-  j= key >> TSIN_HASH_SHIFT;
+  phokey_t pp[MAX_PHRASE_LEN + 1];
+  extract_pho(c_idx, mlen, pp);
 
-  if (j >= TSIN_HASH_N)
+  if (!tsin_seek(pp, 2, &sti, &edi))
     return;
 
-  sti=hashidx[j];
-  edi=hashidx[j+1];
-  phrase_count = 0;
+  while (sti < edi && phrase_count < phkbm.selkeyN) {
+    phokey_t stk[MAX_PHRASE_LEN];
+    usecount_t usecount;
+    u_char stch[MAX_PHRASE_LEN * CH_SZ + 1];
 
-  while (sti < edi && phrase_count < MAX_PHRASE_SEL_N) {
-    char usecount=0;
     load_tsin_entry(sti, &len, &usecount, stk, stch);
 
-    if (len > mlen) {
+    if (len > mlen || len==1) {
       sti++;
       continue;
     }
 
-
     if (chpho_eq_pho(c_idx, stk, len)) {
       sellen[phrase_count]=len;
-      selidx[phrase_count]=sti;
       utf8cpyN(selstr[phrase_count++], stch, len);
     }
 
@@ -580,8 +587,6 @@ static void get_sel_phrase()
   }
 }
 
-
-
 static void get_sel_pho()
 {
   phokey_t key;
@@ -646,7 +651,7 @@ static void disp_current_sel_page()
   if (current_page > 0)
     disp_arrow_up();
 
-  disp_selections(c_idx);
+  disp_selections(c_idx==c_len?c_idx-1:c_idx);
 }
 
 static int fetch_user_selection(int val, char **seltext)
@@ -672,7 +677,7 @@ static int fetch_user_selection(int val, char **seltext)
 
 int phokey_t_seq(phokey_t *a, phokey_t *b, int len);
 
-static void extract_pho(int chpho_idx, int plen, phokey_t *pho)
+void extract_pho(int chpho_idx, int plen, phokey_t *pho)
 {
   int i;
 
@@ -682,7 +687,6 @@ static void extract_pho(int chpho_idx, int plen, phokey_t *pho)
 }
 
 
-gboolean tsin_seek(phokey_t *pho, int plen, int *r_sti, int *r_edi);
 void mask_key_typ_pho(phokey_t *key);
 
 static int qcmp_pre_sel_usecount(const void *aa, const void *bb)
@@ -693,6 +697,27 @@ static int qcmp_pre_sel_usecount(const void *aa, const void *bb)
   return b->usecount - a->usecount;
 }
 
+gboolean check_fixed_mismatch(int chpho_idx, char *mtch, int plen)
+{
+  int j;
+  char *p = mtch;
+
+  for(j=0; j < plen; j++) {
+    int u8sz = utf8_sz(p);
+    if (!(chpho[chpho_idx+j].flag & FLAG_CHPHO_FIXED))
+      continue;
+
+    if (memcmp(chpho[chpho_idx+j].ch, p, u8sz))
+      break;
+
+    p+= u8sz;
+  }
+
+  if (j < plen)
+    return TRUE;
+
+  return FALSE;
+}
 
 static u_char scanphr(int chpho_idx, int plen, gboolean pho_incr)
 {
@@ -727,23 +752,28 @@ static u_char scanphr(int chpho_idx, int plen, gboolean pho_incr)
   pre_selN = 0;
   int maxlen=0;
 
-  while (sti < edi && pre_selN < 10) {
+#define selNMax 30
+  PRE_SEL sel[selNMax];
+  int selN = 0;
+
+
+  while (sti < edi && selN < selNMax) {
     phokey_t mtk[MAX_PHRASE_LEN];
     u_char mtch[MAX_PHRASE_LEN*CH_SZ+1];
-    u_char match_len;
-    char usecount;
+    char match_len;
+    usecount_t usecount;
 
     load_tsin_entry(sti, &match_len, &usecount, mtk, mtch);
 
-
     sti++;
     if (plen > match_len || (pho_incr && plen==match_len)) {
       continue;
     }
 
-    int i;
 
-    for(i=0;i < plen;i++) {
+
+    int i;
+    for(i=0; i < plen; i++) {
       if (mtk[i]!=pp[i])
         break;
     }
@@ -776,40 +806,32 @@ static u_char scanphr(int chpho_idx, int plen, gboolean pho_incr)
 #endif
 
 #if 1
-    int j;
-    char *p = mtch;
-
-    for(j=0; j < plen; j++) {
-      int u8sz = utf8_sz(p);
-      if (!(chpho[chpho_idx+j].flag & FLAG_CHPHO_FIXED))
-        continue;
-
-      if (memcmp(chpho[chpho_idx+j].ch, p, u8sz))
-        break;
-
-      p+= u8sz;
-    }
-
-    if (j < plen)
+    if (check_fixed_mismatch(chpho_idx, mtch, plen))
       continue;
 #endif
 
     if (maxlen < match_len)
       maxlen = match_len;
 
-    pre_sel[pre_selN].len = match_len;
-    pre_sel[pre_selN].phidx = sti - 1;
-    pre_sel[pre_selN].usecount = usecount;
-    utf8cpyN(pre_sel[pre_selN].str, mtch, match_len);
-    memcpy(pre_sel[pre_selN].phokey, mtk, match_len*sizeof(phokey_t));
-    pre_selN++;
+    sel[selN].len = match_len;
+    sel[selN].phidx = sti - 1;
+    sel[selN].usecount = usecount;
+    utf8cpyN(sel[selN].str, mtch, match_len);
+    memcpy(sel[selN].phokey, mtk, match_len*sizeof(phokey_t));
+    selN++;
   }
 
-  qsort(pre_sel, pre_selN, sizeof(PRE_SEL), qcmp_pre_sel_usecount);
+  qsort(sel, selN, sizeof(PRE_SEL), qcmp_pre_sel_usecount);
+
+  pre_selN = Min(selN, phkbm.selkeyN);
+
+  memcpy(pre_sel, sel, sizeof(PRE_SEL) * pre_selN);
 
   return maxlen;
 }
 
+void hide_selections_win();
+
 static void disp_pre_sel_page()
 {
 
@@ -819,8 +841,10 @@ static void disp_pre_sel_page()
     return;
   }
 
-  if (pre_selN==0 || (pre_selN==1 && pre_sel[0].len==2))
+  if (pre_selN==0 || (pre_selN==1 && pre_sel[0].len<=2) || ph_sta < 0) {
+    hide_selections_win();
     return;
+  }
 
   clear_sele();
 
@@ -829,37 +853,12 @@ static void disp_pre_sel_page()
 
     set_sele_text(i, pre_sel[i].str, tlen);
   }
-
+#if 0
+  dbg("ph_sta:%d\n", ph_sta);
+#endif
   disp_selections(ph_sta);
 }
 
-
-static void raise_phr(int c)
-{
-  int i,j,tmp;
-  FILE *fp;
-
-  if (!c)
-    return;
-
-  i=selidx[c];
-  j=selidx[0];
-  tmp=phidx[i];
-  phidx[i]=phidx[j];
-  phidx[j]=tmp;
-  selidx[0]=tmp;
-
-  if ((fp=fopen(tsidxfname,"r+"))==NULL) {
-    dbg("%s modify err", tsidxfname);
-    return;
-  }
-
-  fseek(fp,sizeof(phcount)+sizeof(hashidx) + j * sizeof(int), SEEK_SET);
-  fwrite(&phidx[j],4,(c+1),fp);
-  fclose(fp);
-}
-
-void hide_selections_win();
 static void close_selection_win()
 {
   hide_selections_win();
@@ -867,8 +866,6 @@ static void close_selection_win()
   pre_selN = 0;
 }
 
-
-
 void show_button_pho(gboolean bshow);
 
 void tsin_set_eng_ch(int nmod)
@@ -878,12 +875,17 @@ void tsin_set_eng_ch(int nmod)
   drawcursor();
 
   show_button_pho(eng_ph);
-
+#if TRAY_ENABLED
+  load_tray_icon();
+#endif
+#if 0
   show_win0();
+#endif
 }
 
 void tsin_toggle_eng_ch()
 {
+  compact_win0_x();
   tsin_set_eng_ch(!eng_ph);
 }
 
@@ -908,30 +910,31 @@ static void hide_pre_sel()
 }
 
 
-void clear_tsin_line();
-
-static void clear_disp_ph_sta()
-{
-  clear_tsin_line();
-}
 
 void draw_underline(int index);
 
-static void disp_ph_sta()
+static void draw_ul(start, stop)
 {
-  clear_disp_ph_sta();
+  int i;
+  for(i=start; i < stop; i++)
+    draw_underline(i);
+}
+
+void disp_ph_sta_idx(int idx)
+{
+//  dbg("ph_sta:%d\n", ph_sta);
+  clear_tsin_line();
 
   if (ph_sta < 0)
     return;
 
-  int i;
-
-  for(i=ph_sta; i < c_idx; i++) {
-    draw_underline(i);
-  }
+  draw_ul(idx, c_idx);
 }
 
-
+void disp_ph_sta()
+{
+  disp_ph_sta_idx(ph_sta);
+}
 
 void ch_pho_cpy(CHPHO *pchpho, char *utf8, phokey_t *phos, int len)
 {
@@ -975,23 +978,6 @@ gboolean add_to_tsin_buf(char *str, phokey_t *pho, int len)
     if (c_idx < 0 || c_len + len >= MAX_PH_BF_EXT)
       return 0;
 
-
-    char *pp = str;
-#if 0
-    for(i=0; i < len; i++) {
-      if (pho[i])
-        continue;
-
-      phokey_t tpho[32];
-      tpho[0]=0;
-      int u8len = utf8_sz(pp);
-
-      utf8_pho_keys(pp, tpho);
-      pho[i] = tpho[0];
-      pp += u8len;
-    }
-#endif
-
     if (c_idx < c_len) {
       for(i=c_len-1; i >= c_idx; i--) {
         chpho[i+len] = chpho[i];
@@ -1009,15 +995,17 @@ gboolean add_to_tsin_buf(char *str, phokey_t *pho, int len)
     disp_in_area_pho_tsin();
 
     prbuf();
+
+    set_fixed(c_idx, len);
 #if 1
     for(i=1;i < len; i++) {
       chpho[c_idx+i].psta= c_idx;
     }
 #endif
-
-    if (len > 1)
+#if 0
+    if (len > 0)
       chpho[c_idx].flag |= FLAG_CHPHO_PHRASE_HEAD;
-
+#endif
     drawcursor();
     disp_ph_sta();
     hide_pre_sel();
@@ -1029,11 +1017,12 @@ gboolean add_to_tsin_buf(char *str, phokey_t *pho, int len)
     return TRUE;
 }
 
+#if 1
 static void set_phrase_link(int idx, int len)
 {
     int j;
 
-    if (len < 2)
+    if (len < 1)
       return;
 
     for(j=1;j < len; j++) {
@@ -1042,6 +1031,8 @@ static void set_phrase_link(int idx, int len)
 
     chpho[idx].flag |= FLAG_CHPHO_PHRASE_HEAD;
 }
+#endif
+
 
 // should be used only if it is a real phrase
 gboolean add_to_tsin_buf_phsta(char *str, phokey_t *pho, int len)
@@ -1057,18 +1048,32 @@ gboolean add_to_tsin_buf_phsta(char *str, phokey_t *pho, int len)
     if (idx + len >= MAX_PH_BF_EXT)
       flush_tsin_buffer();
 
+    if (c_idx < c_len) {
+      int avlen = c_idx - ph_sta;
+
+      dbg("avlen:%d %d\n", avlen, len);
+      if (avlen < len) {
+        int d = len - avlen;
+
+        memmove(&chpho[c_idx + d], &chpho[c_idx], sizeof(CHPHO) * (c_len - c_idx));
+        c_len += d;
+      }
+    } else
+      c_len = idx + len;
+
     ch_pho_cpy(&chpho[idx], str, pho, len);
     set_chpho_ch2(&chpho[idx], str, len);
+    set_fixed(idx, len);
 
-    c_len=c_idx=idx + len;
+    c_idx=idx + len;
 
     clrin_pho_tsin();
     disp_in_area_pho_tsin();
 
     prbuf();
-
+#if 1
     set_phrase_link(idx, len);
-
+#endif
     drawcursor();
     disp_ph_sta();
     hide_pre_sel();
@@ -1099,6 +1104,23 @@ void add_to_tsin_buf_str(char *str)
   add_to_tsin_buf(str, pho, N);
 }
 
+int tsin_sele_by_idx(int c)
+{
+  int len = pre_sel[c].len;
+
+  if (c >= pre_selN)
+    return 0;
+
+#if 0
+    dbg("eqlenN:%d %d\n", eqlenN, current_ph_idx);
+#endif
+
+  full_match = FALSE;
+  gboolean b_added = add_to_tsin_buf_phsta(pre_sel[c].str, pre_sel[c].phokey, len);
+
+  return b_added;
+}
+
 
 static gboolean pre_sel_handler(KeySym xkey)
 {
@@ -1125,33 +1147,7 @@ static gboolean pre_sel_handler(KeySym xkey)
 
   c = p - phkbm.selkey;
 
-  int len = pre_sel[c].len;
-
-  if (c >= pre_selN)
-    return 0;
-
-  int j, eqlenN=0, current_ph_idx;
-
-  for(j=0; j < pre_selN; j++) {
-    if (pre_sel[j].len != len || phokey_t_seq(pre_sel[j].phokey, pre_sel[c].phokey, len))
-      continue;
-
-    if (j==c)
-      current_ph_idx = eqlenN;
-
-    selidx[eqlenN++]=pre_sel[j].phidx;
-  }
-
-//    dbg("eqlenN:%d %d\n", eqlenN, current_ph_idx);
-
-  if (eqlenN > 1) {
-    raise_phr(current_ph_idx);
-  }
-
-  full_match = FALSE;
-  gboolean b_added = add_to_tsin_buf_phsta(pre_sel[c].str, pre_sel[c].phokey, len);
-
-  return b_added;
+  return tsin_sele_by_idx(c);
 }
 
 
@@ -1193,14 +1189,136 @@ static gboolean pre_punctuation_hsu(KeySym xkey)
 gboolean inph_typ_pho(char newkey);
 
 
-gint64 current_time()
-{
-  struct timeval tval;
+gint64 current_time();
 
-  gettimeofday(&tval, NULL);
-  return (gint64)tval.tv_sec * 1000000 + tval.tv_usec;
+static void call_tsin_parse()
+{
+  TSIN_PARSE parse[MAX_PH_BF_EXT+1];
+  bzero(parse, sizeof(parse));
+  tsin_parse(parse);
+  prbuf();
 }
 
+static KeySym keypad_proc(KeySym xkey)
+{
+  if (xkey <= XK_KP_9 && xkey >= XK_KP_0)
+    xkey=xkey-XK_KP_0+'0';
+  else {
+    switch (xkey) {
+      case XK_KP_Add:
+        xkey = '+';
+        break;
+      case XK_KP_Subtract:
+        xkey = '-';
+        break;
+      case XK_KP_Multiply:
+        xkey = '*';
+        break;
+      case XK_KP_Divide:
+        xkey = '/';
+        break;
+      case XK_KP_Decimal:
+        xkey = '.';
+        break;
+      default:
+        return 0;
+    }
+  }
+
+  return xkey;
+}
+
+static int cursor_left()
+{
+  close_selection_win();
+  if (c_idx) {
+    clrcursor();
+    c_idx--;
+    drawcursor();
+    return 1;
+  }
+  // Thanks to PCMan.bbs@bbs.sayya.org for the suggestion
+  return c_len;
+}
+static int cursor_right()
+{
+  close_selection_win();
+  if (c_idx < c_len) {
+    clrcursor();
+    c_idx++;
+    drawcursor();
+    return 1;
+  }
+
+  // Thanks to PCMan.bbs@bbs.sayya.org for the suggestion
+  return c_len;
+}
+
+static int cursor_delete()
+{
+  if (c_idx == c_len)
+    return 0;
+  close_selection_win();
+  clear_tsin_line();
+  ityp3_pho=0;
+  pre_selN = 0;
+
+  int j;
+  for(j=3;j>=0;j--)
+    if (typ_pho[j]) {
+      typ_pho[j]=0;
+      disp_in_area_pho_tsin();
+      return 1;
+    }
+
+  clrcursor();
+  int k;
+  int pst=k=chpho[c_idx].psta;
+
+  for(k=c_idx;k<c_len;k++) {
+    chpho[k] = chpho[k+1];
+    chpho[k].psta=chpho[k+1].psta-1;
+  }
+
+  c_len--;
+  hide_char(c_len);
+  init_chpho_i(c_len);
+
+  prbuf();
+
+  compact_win0_x();
+
+  if (!c_idx)
+    clear_match();
+  else {
+    k=c_idx-1;
+    pst=chpho[k].psta;
+
+    while (k>0 && chpho[k].psta==pst)
+      k--;
+
+    if (chpho[k].psta!=pst)
+      k++;
+
+    int match_len= c_idx - k;
+    if (!(match_len=scanphr(k, match_len, FALSE)))
+      ph_sta=-1;
+    else
+      ph_sta=k;
+
+    if (ph_sta < 0 || c_idx - ph_sta < 2)
+      pre_selN = 0;
+  }
+
+  if (!c_len && gcin_pop_up_win)
+    hide_win0();
+
+  disp_ph_sta();
+  return 1;
+}
+
+void case_inverse();
+void pho_play(phokey_t key);
 
 int feedkey_pp(KeySym xkey, int kbstate)
 {
@@ -1213,28 +1331,55 @@ int feedkey_pp(KeySym xkey, int kbstate)
   int j,jj,kk, idx;
   char kno;
 
+  if (tsin_chinese_english_toggle_key == TSIN_CHINESE_ENGLISH_TOGGLE_KEY_CapsLock)
+    eng_ph = !(kbstate&LockMask);
+
   key_press_time = 0;
 
    if (kbstate & (Mod1Mask|Mod1Mask)) {
-       return 0;
+     return 0;
    }
 
+   if (xkey==XK_Shift_L||xkey==XK_Shift_R)
+      key_press_time = current_time();
 
+   if (!eng_ph && !c_len && gcin_pop_up_win && xkey!=XK_Caps_Lock) {
+     hide_win0();
+     return 0;
+   }
+
+   int o_sel_pho = sel_pho;
    close_win_pho_near();
 
    switch (xkey) {
      case XK_Escape:
-        tsin_reset_in_pho();
-        return 1;
+       tsin_reset_in_pho0();
+       if (typ_pho_empty()) {
+         if (!c_len)
+           return 0;
+         if (!o_sel_pho && tsin_tab_phrase_end)
+           goto tab_phrase_end;
+       }
+       tsin_reset_in_pho();
+       return 1;
      case XK_Return:
      case XK_KP_Enter:
         if (shift_m) {
+          if (!c_len)
+            return 0;
           save_frm=c_idx;
           save_to=c_len-1;
+          draw_ul(c_idx, c_len);
           save_phrase();
           return 1;
-        } else
-          return flush_tsin_buffer();
+        } else {
+          if (c_len)
+            flush_tsin_buffer();
+          else
+          if (typ_pho_empty())
+            return 0;
+          return 1;
+        }
      case XK_Home:
         close_selection_win();
         if (!c_len)
@@ -1252,31 +1397,9 @@ int feedkey_pp(KeySym xkey, int kbstate)
         drawcursor();
         return 1;
      case XK_Left:
-        close_selection_win();
-        if (c_idx) {
-          clrcursor();
-          c_idx--;
-          drawcursor();
-          return 1;
-        }
-        // Thanks to PCMan.bbs@bbs.sayya.org for the suggestion
-        if (c_len)
-          return 1;
-        return 0;
+        return cursor_left();
      case XK_Right:
-        close_selection_win();
-        if (c_idx < c_len) {
-          clrcursor();
-          c_idx++;
-          drawcursor();
-          return 1;
-        }
-
-        // Thanks to PCMan.bbs@bbs.sayya.org for the suggestion
-        if (c_len)
-          return 1;
-
-        return 0;
+        return cursor_right();
      case XK_Caps_Lock:
         if (tsin_chinese_english_toggle_key == TSIN_CHINESE_ENGLISH_TOGGLE_KEY_CapsLock) {
           close_selection_win();
@@ -1285,14 +1408,19 @@ int feedkey_pp(KeySym xkey, int kbstate)
         } else
           return 0;
      case XK_Tab:
+        close_selection_win();
         if (tsin_chinese_english_toggle_key == TSIN_CHINESE_ENGLISH_TOGGLE_KEY_Tab) {
-          close_selection_win();
           tsin_toggle_eng_ch();
           return 1;
         }
 
-        if (tsin_tab_phrase_end) {
-          ph_sta = -1;
+        if (tsin_tab_phrase_end && c_len > 1) {
+tab_phrase_end:
+          if (c_idx==c_len)
+            chpho[c_idx-1].flag |= FLAG_CHPHO_PHRASE_USER_HEAD;
+          else
+            chpho[c_idx].flag |= FLAG_CHPHO_PHRASE_USER_HEAD;
+           call_tsin_parse();
           return 1;
         } else {
           if (c_len) {
@@ -1301,69 +1429,11 @@ int feedkey_pp(KeySym xkey, int kbstate)
           }
         }
         return 0;
-     case XK_Shift_L:
-     case XK_Shift_R:
-        key_press_time = current_time();
-        return 1;
      case XK_Delete:
-        if (c_idx == c_len)
-          return 0;
-
-        close_selection_win();
-        clear_disp_ph_sta();
-        ityp3_pho=0;
-        pre_selN = 0;
-
-        for(j=3;j>=0;j--)
-          if (typ_pho[j]) {
-            typ_pho[j]=0;
-            disp_in_area_pho_tsin();
-            return 1;
-          }
-
-        clrcursor();
-        pst=k=chpho[c_idx].psta;
-
-        for(k=c_idx;k<c_len;k++) {
-          chpho[k] = chpho[k+1];
-          chpho[k].psta=chpho[k+1].psta-1;
-        }
-
-        c_len--;
-        init_chpho_i(c_len);
-
-        prbuf();
-
-        compact_win0_x();
-
-        if (!c_idx) {
-          clear_match();
-        }
-        else {
-          k=c_idx-1;
-          pst=chpho[k].psta;
-
-          while (k>0 && chpho[k].psta==pst)
-            k--;
-
-          if (chpho[k].psta!=pst)
-            k++;
-
-          match_len= c_idx - k;
-          if (!(match_len=scanphr(k, match_len, FALSE)))
-            ph_sta=-1;
-          else
-            ph_sta=k;
-
-//          if (ph_sta < 0 || c_idx - ph_sta < 2)
-            pre_selN = 0;
-        }
-
-        disp_ph_sta();
-        return 1;
+        return cursor_delete();
      case XK_BackSpace:
         close_selection_win();
-        clear_disp_ph_sta();
+        clear_tsin_line();
         ityp3_pho=0;
         pre_selN = 0;
 
@@ -1376,6 +1446,9 @@ int feedkey_pp(KeySym xkey, int kbstate)
             if (pre_selN > 1 && scanphr(ph_sta, c_idx - ph_sta, TRUE)) {
               disp_pre_sel_page();
             }
+
+            if (!c_len && gcin_pop_up_win && !j)
+              hide_win0();
             return 1;
           }
 
@@ -1392,14 +1465,14 @@ int feedkey_pp(KeySym xkey, int kbstate)
         }
 
         c_len--;
+        hide_char(c_len);
         init_chpho_i(c_len);
         prbuf();
         compact_win0_x();
 
         if (!c_idx) {
           clear_match();
-        }
-        else {
+        } else {
           k=c_idx-1;
           pst=chpho[k].psta;
 
@@ -1420,6 +1493,10 @@ int feedkey_pp(KeySym xkey, int kbstate)
         }
 
         disp_ph_sta();
+
+        if (!c_len && gcin_pop_up_win)
+          hide_win0();
+
         return 1;
      case XK_Up:
        if (!sel_pho) {
@@ -1450,12 +1527,9 @@ int feedkey_pp(KeySym xkey, int kbstate)
          return 1;
        }
 
-       if (tsin_space_opt == TSIN_SPACE_OPT_FLUSH_BUFFER) {
-         if (c_len && eng_ph
-           && (ityp3_pho || (!typ_pho[0] && !typ_pho[1] && !typ_pho[2])) ) {
-           flush_tsin_buffer();
-           return 1;
-         }
+       if (tsin_space_opt == TSIN_SPACE_OPT_INPUT && !typ_pho[0] && !typ_pho[1] && !typ_pho[2] && !ityp3_pho && !sel_pho) {
+         close_selection_win();
+         goto asc_char;
        }
      case XK_Down:
        if (!eng_ph && xkey == XK_space)
@@ -1506,18 +1580,39 @@ other_keys:
        if ((kbstate & ControlMask)) {
          if (xkey==XK_u) {
            clear_tsin_buffer();
+           if (gcin_pop_up_win)
+             hide_win0();
+           return 1;
+         } else if (tsin_buffer_editing_mode && xkey == XK_e) { //ctrl+e only works when user enabled tsin_buffer_editing_mode
+           //toggler
+           tsin_buffer_editing ^= 1;
+           return 1;
+         } else if (xkey>=XK_1 && xkey<=XK_3) {
            return 1;
          } else
            return 0;
        }
 
+       char xkey_lcase = xkey;
+       if ('A' <= xkey && xkey <= 'Z') {
+         xkey_lcase = tolower(xkey);
+       }
+
+       if (tsin_buffer_editing) {
+         if (xkey_lcase=='h')
+           return cursor_left();
+         else
+         if (xkey_lcase=='l')
+           return cursor_right();
+         else
+         if (xkey_lcase=='x')
+           return cursor_delete();
+       }
+
+       if (tsin_space_opt == TSIN_SPACE_OPT_INPUT)
+         xkey_lcase = keypad_proc(xkey);
 
        char *pp;
-
-       char xkey_lcase = xkey;
-       if ('A' <= xkey && xkey <= 'Z')
-          xkey_lcase = tolower(xkey);
-
        if ((pp=strchr(phkbm.selkey,xkey_lcase)) && sel_pho) {
          int c=pp-phkbm.selkey;
          char *sel_text;
@@ -1525,21 +1620,19 @@ other_keys:
 
          if (len > 1) {
            int cpsta = chpho[c_idx].psta;
-           if (cpsta >= 0) {
-             chpho[cpsta].flag |= FLAG_CHPHO_PHRASE_VOID;
+           if (cpsta >= 0)
              set_chpho_ch(&chpho[c_idx], sel_text, len);
-           } else
+           else
              set_chpho_ch2(&chpho[c_idx], sel_text, len);
 
            chpho[c_idx].flag &= ~FLAG_CHPHO_PHRASE_VOID;
-           set_phrase_link(c_idx, len);
-           raise_phr(c);
+           set_fixed(c_idx, len);
+
+           call_tsin_parse();
+
            if (c_idx + len == c_len) {
              ph_sta = -1;
-
-             for(i=c_idx; i < c_len; i++) {
-               draw_underline(i);
-             }
+             draw_ul(c_idx, c_len);
            }
          } else
          if (len == 1) { // single chinese char
@@ -1549,10 +1642,13 @@ other_keys:
 
            if (i && chpho[i].psta == i-1 && !(chpho[i-1].flag & FLAG_CHPHO_FIXED)) {
              set_chpho_ch(&chpho[i-1], chpho[i-1].ph1ch, 1);
+#if 0
              chpho[i-1].flag |= FLAG_CHPHO_PHRASE_VOID;
+#endif
            }
 
-           chpho[i].flag |= FLAG_CHPHO_FIXED;
+           set_fixed(i, 1);
+           call_tsin_parse();
          }
 
          if (len) {
@@ -1574,13 +1670,18 @@ other_keys:
        sel_pho=current_page=0;
    }
 
-   if (!eng_ph || shift_m || (xkey <= XK_KP_9 && xkey >= XK_KP_0) ||
-        xkey == XK_KP_Subtract || xkey == XK_KP_Add || xkey == XK_KP_Multiply ||
-        xkey == XK_KP_Divide
-       ) {
+   KeySym key_pad = keypad_proc(xkey);
+
+   if (xkey > 0x7e && !key_pad)
+     return 0;
+
+   if (!eng_ph || typ_pho[0]!=BACK_QUOTE_NO && (shift_m || key_pad || !phkbm.phokbm[xkey][0].num)) {
+       if (key_pad)
+         xkey = key_pad;
 asc_char:
         if (shift_m) {
           if (pre_sel_handler(xkey)) {
+            call_tsin_parse();
             return 1;
           }
 
@@ -1588,24 +1689,6 @@ asc_char:
             return 1;
         }
 
-        if (xkey <= XK_KP_9 && xkey >= XK_KP_0)
-          xkey=xkey-XK_KP_0+'0';
-        else {
-          switch (xkey) {
-            case XK_KP_Add:
-              xkey = '+';
-              break;
-            case XK_KP_Subtract:
-              xkey = '-';
-              break;
-            case XK_KP_Multiply:
-              xkey = '*';
-              break;
-            case XK_KP_Divide:
-              xkey = '/';
-              break;
-          }
-        }
 
 //        dbg("xkey: %c\n", xkey);
 
@@ -1614,25 +1697,23 @@ asc_char:
 
           if (!(kbstate&LockMask) && ppp && !((ppp-ochars) & 1))
             xkey=*(ppp+1);
+
+#if 0
           if (kbstate&LockMask && islower(xkey))
             xkey-=0x20;
           else
             if (!(kbstate&LockMask) && isupper(xkey))
               xkey+=0x20;
+#endif
         } else {
-          if (!eng_ph && tsin_chinese_english_toggle_key == TSIN_CHINESE_ENGLISH_TOGGLE_KEY_CapsLock) {
-            if (shift_m) {
-              if (islower(xkey))
-                xkey-=0x20;
-            } else
-            if (isupper(xkey))
-              xkey+=0x20;
+          if (!eng_ph && tsin_chinese_english_toggle_key == TSIN_CHINESE_ENGLISH_TOGGLE_KEY_CapsLock
+              && gcin_capslock_lower) {
+            case_inverse(&xkey, shift_m);
           }
         }
 
-
         if (xkey > 127)
-          return 1;
+          return 0;
 
         u_char tt=xkey;
         shift_ins();
@@ -1644,6 +1725,7 @@ asc_char:
           chpho[c_idx].ch[0]=tt;
         }
 
+        set_fixed(c_idx, 1);
         phokey_t tphokeys[32];
         tphokeys[0]=0;
         utf8_pho_keys(chpho[c_idx].ch, tphokeys);
@@ -1654,23 +1736,23 @@ asc_char:
         if (c_idx < c_len)
           prbuf();
 
-        orig_disp_in_area_x = -1;
-
         if (gcin_pop_up_win)
           show_win0();
 
         drawcursor();
         return 1;
    } else { /* pho */
-     if (xkey > 127)
+     if (xkey > 127) {
        return 0;
+     }
 
      // for hsu & et26
-     if (strchr(hsu_punc, xkey) && !phkbm.phokbm[xkey][0].num)
+     if (strchr(hsu_punc, xkey) && !phkbm.phokbm[xkey][0].num && typ_pho[0]!=BACK_QUOTE_NO)
        return pre_punctuation_hsu(xkey);
 
-     if (xkey >= 'A' && xkey <='Z')
+     if (xkey >= 'A' && xkey <='Z' && typ_pho[0]!=BACK_QUOTE_NO)
        xkey+=0x20;
+//     printf("bbbb %c\n", xkey);
 
      inph_typ_pho(xkey);
 
@@ -1695,25 +1777,33 @@ llll2:
 
      disp_in_area_pho_tsin();
 
-     if (pre_selN && scanphr(ph_sta, c_idx - ph_sta, TRUE)) {
-       disp_pre_sel_page();
-     }
+     if (pre_selN)
+       scanphr(ph_sta, c_idx - ph_sta, TRUE);
+
+     disp_pre_sel_page();
 
      key = pho2key(typ_pho);
 
+     pho_play(key);
+
      int vv=hash_pho[typ_pho[0]];
+
      phokey_t ttt=0xffff;
      while (vv<idxnum_pho) {
        ttt=idx_pho[vv].key;
-       if (!typ_pho[0]) ttt &= ~(31<<9);
-       if (!typ_pho[1]) ttt &= ~(3<<7);
-       if (!typ_pho[2]) ttt &= ~(15<<3);
-       if (!typ_pho[3]) ttt &= ~(7);
+       if (typ_pho[0]!=BACK_QUOTE_NO) {
+         if (!typ_pho[0]) ttt &= ~(31<<9);
+         if (!typ_pho[1]) ttt &= ~(3<<7);
+         if (!typ_pho[2]) ttt &= ~(15<<3);
+         if (!typ_pho[3]) ttt &= ~(7);
+       }
        if (ttt>=key) break;
        else
        vv++;
      }
-
+#if 0
+     printf("aaaa vv:%d  idxnum_pho:%d   ttt:%x key:%x\n",vv, idxnum_pho, ttt, key);
+#endif
      if (ttt > key || (ityp3_pho && idx_pho[vv].key!=key) ) {
        while (jj<4) {
          while(kk<3)
@@ -1737,21 +1827,33 @@ llll2:
        return 1;
      }
 
-     if (key==0 || !ityp3_pho)
+     if (typ_pho[0]==BACK_QUOTE_NO && typ_pho[1])
+       ityp3_pho = 1;
+
+     if (key==0 || !ityp3_pho) {
        return 1;
+     }
 
      ii=idx_pho[vv].start;
      start_idx=ii;
+#if 0
+     printf("%x %x %d vv:%d idxnum_pho:%d-->", ttt, key, start_idx, vv, idxnum_pho);
+     utf8_putchar(ch_pho[start_idx].ch);
+     puts("<---");
+#endif
    } /* pho */
 
    put_b5_char(ch_pho[start_idx].ch, key);
 
+   call_tsin_parse();
+
    disp_ph_sta();
    clrin_pho_tsin();
    clr_in_area_pho_tsin();
    drawcursor();
    hide_pre_sel();
 
+
    if (ph_sta < 0) {
 restart:
      if ((match_len=scanphr(c_idx-1,1, FALSE)))
@@ -1844,11 +1946,11 @@ restart:
 
        if (j < mdist)
          continue;
-
+#if 0
        ch_pho_cpy(&chpho[ph_sta], pre_sel[i].str, pre_sel[i].phokey, mdist);
        if (chpho[ph_sta].psta < 0)
          set_chpho_ch2(&chpho[ph_sta], pre_sel[i].str, mdist);
-
+#endif
 
        int j;
        for(j=0;j < mdist; j++) {
@@ -1858,8 +1960,9 @@ restart:
        }
 
        full_match = TRUE;
+#if 0
        chpho[ph_sta].flag |= FLAG_CHPHO_PHRASE_HEAD;
-
+#endif
        if (mdist==max_match_phrase_len) { // tstr: 選擇視窗
 //         dbg("full match .......... %d\n", ph_sta);
          ph_sta_last = ph_sta;
@@ -1881,8 +1984,13 @@ int feedkey_pp_release(KeySym xkey, int kbstate)
   switch (xkey) {
      case XK_Shift_L:
      case XK_Shift_R:
-        if (tsin_chinese_english_toggle_key == TSIN_CHINESE_ENGLISH_TOGGLE_KEY_Shift &&
-            current_time() - key_press_time < 300000) {
+        if (
+(  (tsin_chinese_english_toggle_key == TSIN_CHINESE_ENGLISH_TOGGLE_KEY_Shift) ||
+   (tsin_chinese_english_toggle_key == TSIN_CHINESE_ENGLISH_TOGGLE_KEY_ShiftL
+     && xkey == XK_Shift_L) ||
+   (tsin_chinese_english_toggle_key == TSIN_CHINESE_ENGLISH_TOGGLE_KEY_ShiftR
+     && xkey == XK_Shift_R))
+          &&  current_time() - key_press_time < 300000) {
           key_press_time = 0;
           close_selection_win();
           tsin_toggle_eng_ch();
